import logging
l = logging.getLogger("claripy.frontends.composite_frontend")

import weakref
import itertools
symbolic_count = itertools.count()

from .constrained_frontend import ConstrainedFrontend
from claripy.ast.strings import String

class CompositeFrontend(ConstrainedFrontend):
    def __init__(self, template_frontend, template_frontend_string, track=False, **kwargs):
        super().__init__(**kwargs)
        self._solvers = { }
        self._unchecked_solvers = weakref.WeakSet()
        self._owned_solvers = weakref.WeakSet()
        self._template_frontend = template_frontend
        self._template_frontend_string = template_frontend_string
        self._unsat = False
        self._track = track

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._unchecked_solvers = weakref.WeakSet()
        c._owned_solvers = weakref.WeakSet()
        c._solvers = { }
        c._template_frontend = self._template_frontend
        if hasattr(self, '_template_frontend_string'):
            c._template_frontend_string = self._template_frontend_string
        c._unsat = False
        c._track = self._track

    def _copy(self, c):
        super()._copy(c)
        c._unsat = self._unsat
        c._track = self._track

        c._solvers = dict(self._solvers)
        c._unchecked_solvers = weakref.WeakSet(self._unchecked_solvers)
        self._owned_solvers = weakref.WeakSet() # for the COW
        return c


    #
    # Serialization stuff
    #

    def __getstate__(self):
        return self._solvers, self._template_frontend, self._unsat, self._track, super().__getstate__()

    def __setstate__(self, s):
        self._solvers, self._template_frontend, self._unsat, self._track, base_state = s
        self._owned_solvers = weakref.WeakSet(self._solver_list)
        self._unchecked_solvers = weakref.WeakSet()
        super().__setstate__(base_state)

    def downsize(self):
        for e in self._solver_list:
            e.downsize()

    #
    # Frontend management
    #

    @property
    def _solver_list(self):
        seen_solvers = set()
        solver_list = [ ]
        for s in self._solvers.values():
            if id(s) in seen_solvers: continue
            seen_solvers.add(id(s))
            solver_list.append(s)
        return solver_list

    @property
    def variables(self):
        return set(self._solvers.keys())

    # this is really hacky, but we want to avoid having our variables messed with
    @variables.setter
    def variables(self, v):
        pass

    #
    # Solver list management
    #

    def _solvers_for_variables(self, names):
        seen_solvers = set()
        existing_solvers = [ ]
        for n in names:
            if n not in self._solvers: continue
            s = self._solvers[n]

            if id(s) in seen_solvers: continue
            seen_solvers.add(id(s))
            existing_solvers.append(s)
        return existing_solvers

    @staticmethod
    def _names_for(names=None, lst=None, lst2=None, e=None, v=None):
        if names is None:
            names = set()
        if e is not None and isinstance(e, Base):
            names.update(e.variables)
        if v is not None and isinstance(v, Base):
            names.update(v.variables)
        if lst is not None:
            for ee in lst:
                if isinstance(ee, Base):
                    names.update(ee.variables)
        if lst2 is not None:
            for ee in lst2:
                if isinstance(ee, Base):
                    names.update(ee.variables)
        return names

    def _merged_solver_for(self, *args, **kwargs):
        return self._solver_for_names(self._names_for(*args, **kwargs))

    def _solver_for_names(self, names):
        l.debug("composite_solver._merged_solver_for() running with %d names", len(names))
        solvers = self._solvers_for_variables(names)
        if len(solvers) == 0:
            if any(var for var in names if var.startswith(String.STRING_TYPE_IDENTIFIER)):
                l.debug("... creating new solver for strings")
                return self._template_frontend_string.blank_copy()
            else:
                l.debug("... creating new solver")
                return self._template_frontend.blank_copy()
        elif len(solvers) == 1:
            l.debug("... got one solver")
            return solvers[0]
        else:
            l.debug(".... combining %d solvers", len(solvers))
            return solvers[0].combine(solvers[1:])

    def _shared_solvers(self, others):
        """
        Returns a sequence of the solvers that self and others share.
        """

        solvers_by_id = { id(s): s for s in self._solver_list }
        common_solvers = set(solvers_by_id.keys())
        other_sets = [ { id(s) for s in cs._solver_list } for cs in others ]
        for o in other_sets: common_solvers &= o

        return [ solvers_by_id[s] for s in common_solvers ]

    def _variable_sets(self):
        return { s.variables for s in self._solver_list }

    def _shared_varsets(self, others):
        common_varsets = self._variable_sets()
        for o in others: common_varsets &= o.all_varsets()
        return common_varsets

    def _split_child(self, s):
        ss = s.split()
        if len(ss) == 1:
            return [ s ]

        l.debug("... split solver %r into %d parts", s, len(ss))
        l.debug("... variable counts: %s", [ len(cs.variables) for cs in ss ])

        for ns in ss:
            self._owned_solvers.add(ns)
            self._store_child(ns)

        return ss

    def _reabsorb_solver(self, s):
        try:
            if len(s.variables) == 0 or self._solvers[min(iter(s.variables))] is s:
                return
        except KeyError:
            # this happens when a variable is introduced due to constraint expansion
            return

        if isinstance(s, ModelCacheMixin):
            new_solvers = s.split()
            old_solvers = self._solvers_for_variables(s.variables)
            if len(new_solvers) == len(old_solvers):
                done = set()
                for ss in new_solvers:
                    if ss in done:
                        continue
                    done.add(ss)
                    v = min(iter(ss.variables))
                    self._solvers[v].update(ss)
            else:
                for ns in new_solvers:
                    self._owned_solvers.add(ns)
                    self._store_child(ns)

    def _store_child(self, ns, extra_names=frozenset(), invalidate_cache=True):
        for v in ns.variables | extra_names:
            #os = self._solvers[v]
            self._solvers[v] = ns
        if invalidate_cache:
            self._unchecked_solvers.add(ns)

        #if isinstance(s, ModelCacheMixin):
        #   if len(os._models) < len(ns._models):
        #       print("GOT %d NEW MODELS (before: %d)" % (
        #           len(ns._models) - len(os._models), len(os._models)
        #       ))
        #   elif len(os._models) > len(ns._models):
        #       print("WARNING: LOST %d NEW MODELS (before: %d)" % (
        #           len(os._models) - len(ns._models), len(os._models)
        #       ))
        #   else:
        #       print("Remained at %d models." % len(os._models))

    #
    # Constraints
    #

    def _claim(self, s):
        if s not in self._owned_solvers:
            sc = s.branch()
            self._owned_solvers.add(sc)
            return sc
        else:
            return s

    def _add_dependent_constraints(self, names, constraints, invalidate_cache=True, **kwargs):
        if not invalidate_cache and len(self._solvers_for_variables(names)) > 1:
            l.debug("Ignoring cross-solver helper constraints.")
            return [ ]

        l.debug("Adding %d constraints to %d names", len(constraints), len(names))
        s = self._claim(self._merged_solver_for(names=names))
        added = s.add(constraints, invalidate_cache=invalidate_cache, **kwargs)
        self._store_child(s, invalidate_cache=invalidate_cache)
        return added

    def add(self, constraints, **kwargs): #pylint:disable=arguments-differ
        split = self._split_constraints(constraints)
        child_added = [ ]

        #l.debug("%s, solvers before: %d", self, len(self._solvers))
        unsure = [ ]
        for names,set_constraints in split:
            if names == { 'CONCRETE' }:
                try:
                    if any(backends.concrete.convert(c) is False for c in set_constraints):
                        self._unsat = True
                except BackendError:
                    unsure.extend(set_constraints)
            else:
                child_added += self._add_dependent_constraints(names, set_constraints, **kwargs)

        #l.debug("... solvers after add: %d", len(self._solver_list))

        if len(unsure) > 0:
            for s in self._solver_list:
                s = self._claim(s)
                s.add(unsure)
                self._store_child(s)

        return super().add(child_added)

    #
    # Solving
    #

    def _ensure_sat(self, extra_constraints):
        if self._unsat or (len(extra_constraints) == 0 and not self.satisfiable()):
            raise UnsatError("CompositeSolver is already unsat")

    def check_satisfiability(self, extra_constraints=(), exact=None):
        if self._unsat:
            return 'UNSAT'

        l.debug("%r checking satisfiability...", self)

        if len(extra_constraints) != 0:
            extra_solver = self._merged_solver_for(lst=extra_constraints)
            extra_solver_satness = extra_solver.check_satisfiability(extra_constraints=extra_constraints, exact=exact)
            if extra_solver_satness in {'UNSAT', 'UNKNOWN'}:
                return extra_solver_satness
            self._reabsorb_solver(extra_solver)

        for s in self._unchecked_solvers:
            if extra_constraints and s.variables & extra_solver.variables:
                # skip solvers covered by extra constraints (they were checked above)
                continue

            if self._solvers[min(iter(s.variables))] is not s:
                # this happens when a parent solver didn't check all unchecked solvers, and we have stale
                # child solvers in the unchecked list
                continue

            satness = s.check_satisfiability(exact=exact)
            if satness in {'UNSAT', 'UNKNOWN'}:
                return satness

        self._unchecked_solvers.clear()
        return 'SAT'

    def satisfiable(self, extra_constraints=(), exact=None):
        return self.check_satisfiability(extra_constraints=extra_constraints, exact=exact) == 'SAT'

    def eval(self, e, n, extra_constraints=(), exact=None):
        self._ensure_sat(extra_constraints=extra_constraints)

        ms = self._merged_solver_for(e=e, lst=extra_constraints)
        r = ms.eval(e, n, extra_constraints=extra_constraints, exact=exact)
        self._reabsorb_solver(ms)
        return r

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        self._ensure_sat(extra_constraints=extra_constraints)

        ms = self._merged_solver_for(lst2=exprs, lst=extra_constraints)
        r = ms.batch_eval(exprs, n, extra_constraints=extra_constraints, exact=exact)
        self._reabsorb_solver(ms)
        return r

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        self._ensure_sat(extra_constraints=extra_constraints)

        ms = self._merged_solver_for(e=e, lst=extra_constraints)
        r = ms.max(e, extra_constraints=extra_constraints, signed=signed, exact=exact)
        self._reabsorb_solver(ms)
        return r

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        self._ensure_sat(extra_constraints=extra_constraints)

        ms = self._merged_solver_for(e=e, lst=extra_constraints)
        r = ms.min(e, extra_constraints=extra_constraints, signed=signed, exact=exact)
        self._reabsorb_solver(ms)
        return r

    def solution(self, e, v, extra_constraints=(), exact=None):
        self._ensure_sat(extra_constraints=extra_constraints)

        ms = self._merged_solver_for(e=e, v=v, lst=extra_constraints)
        r = ms.solution(e, v, extra_constraints=extra_constraints, exact=exact)
        self._reabsorb_solver(ms)
        return r

    def is_true(self, e, extra_constraints=(), exact=None):
        #self._ensure_sat(extra_constraints=extra_constraints)

        ms = self._merged_solver_for(e=e, lst=extra_constraints)
        r = ms.is_true(e, extra_constraints=extra_constraints, exact=exact)
        #self._reabsorb_solver(ms)
        return r

    def is_false(self, e, extra_constraints=(), exact=None):
        #self._ensure_sat(extra_constraints=extra_constraints)

        ms = self._merged_solver_for(e=e, lst=extra_constraints)
        r = ms.is_false(e, extra_constraints=extra_constraints, exact=exact)
        #self._reabsorb_solver(ms)
        return r

    def unsat_core(self, extra_constraints=()):
        if self.satisfiable(extra_constraints=extra_constraints):
            return tuple()

        cores = [ ]

        for solver in self._solver_list:
            cores.extend(list(solver.unsat_core(extra_constraints=extra_constraints)))

        return cores

    def simplify(self):
        if self._unsat:
            return self.constraints

        new_constraints = [ ]

        l.debug("Simplifying %r with %d solvers", self, len(self._solver_list))
        for s in self._solver_list:
            if isinstance(s, SimplifySkipperMixin) and s._simplified:
                new_constraints += s.constraints
                continue

            l.debug("... simplifying child solver %r", s)
            s.simplify()
            results = self._split_child(s)
            for ns in results:
                if isinstance(ns, SimplifySkipperMixin):
                    ns._simplified = True
            new_constraints += s.constraints

        l.debug("... after-split, %r has %d solvers", self, len(self._solver_list))

        self.constraints = new_constraints
        return new_constraints

    #
    # Merging and splitting
    #

    def finalize(self):
        for s in self._solver_list:
            s.finalize()

    @property
    def timeout(self):
        return self._template_frontend.timeout

    @timeout.setter
    def timeout(self, t):
        self._template_frontend.timeout = t
        for s in self._solver_list:
            s.timeout = t

    @staticmethod
    def _merge_with_ancestor(common_ancestor, merge_conditions):
        merged = common_ancestor.branch()
        merged.add([Or(*merge_conditions)])
        #import ipdb; ipdb.set_trace()
        return True, merged

    def merge(self, others, merge_conditions, common_ancestor=None):
        if common_ancestor is not None:
            return self._merge_with_ancestor(common_ancestor, merge_conditions)

        l.debug("Merging %s with %d other solvers.", self, len(others))
        merged = self.blank_copy()
        common_solvers = self._shared_solvers(others)
        common_ids = { id(s) for s in common_solvers }
        l.debug("... %s common solvers", len(common_solvers))

        for s in common_solvers:
            self._owned_solvers.discard(s)
            for o in others:
                o._owned_solvers.discard(s)

            for v in s.variables:
                merged._solvers[v] = s

        noncommon_solvers = [ [ s for s in cs._solver_list if id(s) not in common_ids ] for cs in [self]+others ]

        l.debug("... merging noncommon solvers")
        combined_noncommons = [ ]
        for ns in noncommon_solvers:
            l.debug("... %d", len(ns))
            if len(ns) == 0:
                pass
            elif len(ns) == 1:
                combined_noncommons.append(ns[0])
            else:
                combined_noncommons.append(ns[0].combine(ns[1:]))

        if len(combined_noncommons):
            _, merged_noncommon = combined_noncommons[0].merge(
                combined_noncommons[1:], merge_conditions
            )

            merged._owned_solvers.add(merged_noncommon)
            merged._store_child(merged_noncommon)

        merged.constraints = list(
            itertools.chain.from_iterable(a.constraints for a in merged._solver_list)
        )
        return True, merged

    def combine(self, others):
        combined = self.blank_copy()
        combined.add(self.constraints)
        for o in others:
            combined.add(o.constraints)
        return combined

    def split(self):
        return [ s.branch() for s in self._solver_list ]

from ..ast import Base
from ..ast.bool import Or
from .. import backends
from ..errors import BackendError, UnsatError
from ..frontend_mixins.model_cache_mixin import ModelCacheMixin
from ..frontend_mixins.simplify_skipper_mixin import SimplifySkipperMixin



import subprocess
import logging

import re
from . import SMTLibSolverBackend, PopenSolverProxy
from ...errors import MissingSolverError

log = logging.getLogger(__name__)

def get_version():
    try:
        version_string = subprocess.check_output(['cvc4', '--version']).decode('utf-8')
        version_match = re.match('This is CVC4 version (.*)\n', version_string)

        if not version_match:
            return False, None, "Found malformed version string: {}".format(version_string)

        return True, version_match.group(1), None

    except subprocess.CalledProcessError as ex:
        return False, None, "Not found, error: {}".format(ex)
    except OSError as ex:
        return False, None, "Not found, error: {}".format(ex)


IS_INSTALLED, VERSION, ERROR = get_version()

class CVC4Proxy(PopenSolverProxy):
    def __init__(self, timeout=None):
        # lazy instantiation: Here we don't spawn the subprocess
        self.timeout = timeout
        self.installed = False
        p = None
        super(CVC4Proxy, self).__init__(p)

    def create_process(self):
        # spawn the subprocess
        if not IS_INSTALLED:
            raise MissingSolverError('CVC4 not found! Please install CVC4 before using this backend')
        cmd = ['cvc4', '--lang=smt', '-q', '--strings-exp']
        if self.timeout is not None:
            cmd += ['--tlimit-per={}'.format(self.timeout)]
        p = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.installed = True
        return p

class SolverBackendCVC4(SMTLibSolverBackend):
    def solver(self, timeout=None):
        """
        This function should return an instance of whatever object handles
        solving for this backend. For example, in Z3, this would be z3.Solver().
        """
        return CVC4Proxy(timeout)

from ... import backend_manager as backend_manager
backend_manager.backends._register_backend(SolverBackendCVC4(), 'smtlib_cvc4', False, False)



import hashlib

import os

from claripy.ast.bv import BV

from .. import BackendError
from ..backend_smtlib import BackendSMTLibBase
from ...smtlib_utils import SMTParser, make_pysmt_const_from_type
from six.moves import cStringIO

from pysmt.smtlib.parser import Tokenizer
from pysmt.shortcuts import NotEquals


class AbstractSMTLibSolverProxy(object):
    def write(self, smt):
        raise NotImplementedError

    def read(self, n):
        raise NotImplementedError

    def setup(self):
        pass

    def reset(self):
        self.write('(reset)\n')

    def readuntil(self, s):
        buf = b''
        s = s.encode()
        while s not in buf:
            buf += self.read(1)
        return buf

    def readline(self):
        return self.readuntil('\n')

    def writeline(self, l):
        return self.write(l + '\n')

    def read_sat(self):
        return self.readline().strip().decode('utf-8')

    def read_model(self):
        read_model = self.readuntil('\n)\n').strip().decode('utf-8')
        return read_model

    def create_process(self):
        raise NotImplementedError

class PopenSolverProxy(AbstractSMTLibSolverProxy):
    def __init__(self, p):
        super(PopenSolverProxy, self).__init__()
        self.p = p
        self.constraints = []

    def read(self, n):
        if self.p is None:
            self.p = self.create_process()
        return self.p.stdout.read(n)

    def write(self, smt):
        if self.p is None:
            self.p = self.create_process()
        self.p.stdin.write(smt.encode())
        self.p.stdin.flush()

    def add_constraints(self, csts, track=False):
        self.constraints.extend(csts)

    def terminate(self):
        self.p.terminate()
        self.p = None


class SMTLibSolverBackend(BackendSMTLibBase):
    def __init__(self, *args, **kwargs):
        kwargs['solver_required'] = True
        self.smt_script_log_dir = kwargs.pop('smt_script_log_dir', None)
        super(SMTLibSolverBackend, self).__init__(*args, **kwargs)

    def solver(self, timeout=None): #pylint:disable=no-self-use,unused-argument
        """
        This function should return an instance of whatever object handles
        solving for this backend. For example, in Z3, this would be z3.Solver().
        """
        raise NotImplementedError

    def _get_primitive_for_expr(self, model, e):
        substituted = e.substitute(model).simplify()
        if not substituted.is_constant():
            raise BackendError(
                "CVC4 backend currently only supports requests for symbols directly! This is a weird one that doesn't "
                "turn constant after substitution??")

        return substituted.constant_value()

    def _simplify(self, e):
        return e

    def _is_false(self, e, extra_constraints=(), solver=None, model_callback=None):
        if e.is_constant() and e.constant_value() is False:
            return True
        return False

    def _is_true(self, e, extra_constraints=(), solver=None, model_callback=None):
        if e.is_constant() and e.constant_value() is True:
            return True
        return False

    def _batch_eval(self, exprs, n, extra_constraints=(), solver=None, model_callback=None):
        return [self._eval(e, n, extra_constraints=extra_constraints, solver=solver, model_callback=model_callback) for e in exprs]

    def _add(self, s, c, track=False):
        s.add_constraints(c, track=track)

    def _check_satness(self, solver=None, extra_constraints=(), model_callback=None, extra_variables=()):
        vars, csts = self._get_all_vars_and_constraints(solver=solver, e_c=extra_constraints, e_v=extra_variables)
        smt_script = self._get_satisfiability_smt_script(constraints=csts, variables=vars)

        if self.smt_script_log_dir is not None:
            fname = 'check-sat_{}.smt2'.format(hashlib.md5(smt_script.encode()).hexdigest())

            with open(os.path.join(self.smt_script_log_dir, fname), 'wb') as f:
                f.write(smt_script.encode())

        solver.reset()
        solver.write(smt_script)

        sat = solver.read_sat().upper()
        if sat not in {'SAT', 'UNSAT', 'UNKNOWN'}:
            raise ValueError("Solver error, don't understand (check-sat) response: {}".format(repr(sat)))
        return sat.upper()

    def _check_satisfiability(self, extra_constraints=(), solver=None, model_callback=None, extra_variables=()):
        satness = self._check_satness(solver, extra_constraints, model_callback, extra_variables)
        return satness

    def _satisfiable(self, solver=None, extra_constraints=(), model_callback=None, extra_variables=()):
        satness = self._check_satness(solver, extra_constraints, model_callback, extra_variables)
        # solver is done, terminate process
        solver.terminate()

        return satness == 'SAT'

    def _get_model(self, solver=None, extra_constraints=(), extra_variables=()):
        vars, csts = self._get_all_vars_and_constraints(solver=solver, e_c=extra_constraints, e_v=extra_variables)
        smt_script = self._get_full_model_smt_script(constraints=csts, variables=vars)
        if self.smt_script_log_dir is not None:
            fname = 'get-model_{}.smt2'.format(hashlib.md5(smt_script.encode()).hexdigest())
            with open(os.path.join(self.smt_script_log_dir, fname), 'wb') as f:
                f.write(smt_script.encode())

        solver.reset()
        solver.write(smt_script)

        sat = solver.read_sat()
        if sat == 'sat':
            model_string = solver.read_model()
            tokens = Tokenizer(cStringIO(model_string), interactive=True)
            ass_list = SMTParser(tokens).consume_assignment_list()
            return sat, {s: val for s, val in ass_list}, ass_list
        else:
            error = solver.readline()

        return sat, error, None

    def _eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
        e_c = list(extra_constraints)

        if expr.is_constant():
            return [expr.constant_value()]

        expr_vars = expr.get_free_variables()

        results = []
        while len(results) < n:
            sat, model, ass_list = self._get_model(solver=solver, extra_constraints=e_c, extra_variables=expr_vars)
            if sat != 'sat':
                break

            val = self._get_primitive_for_expr(model, expr)

            if val in results:
                # import ipdb; ipdb.set_trace()
                raise ValueError("Solver error, solver returned the same value twice incorrectly!")

            results.append(val)
            # e_c.append(And(*[NotEquals(s, val) for s, val in ass_list]))
            e_c.append(NotEquals(make_pysmt_const_from_type(val, expr.get_type()), expr))

        return tuple(results)

    def eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
        """
        This function returns up to `n` possible solutions for expression `expr`.

        :param expr: expression (an AST) to evaluate
        :param n: number of results to return
        :param solver: a solver object, native to the backend, to assist in
                       the evaluation (for example, a z3.Solver)
        :param extra_constraints: extra constraints (as ASTs) to add to the solver for this solve
        :param model_callback:      a function that will be executed with recovered models (if any)
        :return:              A sequence of up to n results (backend objects)
        """
        if self._solver_required and solver is None:
            raise BackendError("%s requires a solver for evaluation" % self.__class__.__name__)

        results = self._eval(
            self.convert(expr), n, extra_constraints=self.convert_list(extra_constraints),
            solver=solver, model_callback=model_callback
        )

        results = list(results)
        if type(expr) is not BV:
            return results

        size = expr.length
        for i in range(len(results)):
            results[i] &= (1 << size) - 1 # convert it back to unsigned

        # solver is done, terminate process
        solver.terminate()

        return results

from . import cvc4_popen
from . import z3_popen
from . import abc_popen
from . import z3str_popen



import json

import pysmt
from pysmt.shortcuts import Symbol, get_env
from pysmt.smtlib.parser import SmtLibParser, PysmtSyntaxError


def make_pysmt_const_from_type(val, type):
    return getattr(pysmt.shortcuts, str(type))(val)


class SMTParser:
    def __init__(self, tokens):
        self.p = SmtLibParser()
        self.tokens = tokens

    def expect(self, *allowed):
        t = self.tokens.consume()
        if t not in allowed:
            raise PysmtSyntaxError("Invalid token, expected any of {}, got '{}'".format(allowed, t))
        return t

    def expect_assignment_tuple(self):
        self.expect('(')
        cmd = self.expect('define-fun')
        vname = self.p.parse_atom(self.tokens, cmd)
        self.expect('(')
        self.expect(')')
        t = self.p.parse_type(self.tokens, cmd)
        value = self.p.get_expression(self.tokens)
        self.expect(')')

        return Symbol(vname, t), getattr(pysmt.shortcuts, t.name)(value.constant_value())

    def consume_assignment_list(self):
        self.expect('(')
        self.expect('model')
        """Parses a list of expressions from the tokens"""

        assignments = []
        while True:
            next_token = self.tokens.consume()
            self.tokens.add_extra_token(next_token)  # push it back
            if next_token == ')':
                break

            assignments.append(self.expect_assignment_tuple())

        self.expect(')')

        return assignments


# pylint:disable=isinstance-second-argument-not-valid-type
import collections
import itertools
import operator
from typing import Optional

from functools import reduce


class SimplificationManager:
    def __init__(self):
        self._simplifiers = {
            'Reverse': self.bv_reverse_simplifier,
            'And': self.boolean_and_simplifier,
            'Or': self.boolean_or_simplifier,
            'Not': self.boolean_not_simplifier,
            'Extract': self.extract_simplifier,
            'Concat': self.concat_simplifier,
            'If': self.if_simplifier,
            '__lshift__': self.lshift_simplifier,
            '__rshift__': self.rshift_simplifier,
            'LShR': self.lshr_simplifier,
            '__eq__': self.eq_simplifier,
            '__ne__': self.ne_simplifier,
            '__or__': self.bitwise_or_simplifier,
            '__and__': self.bitwise_and_simplifier,
            '__xor__': self.bitwise_xor_simplifier,
            '__add__': self.bitwise_add_simplifier,
            '__sub__': self.bitwise_sub_simplifier,
            '__mul__': self.bitwise_mul_simplifier,
            'ZeroExt': self.zeroext_simplifier,
            'SignExt': self.signext_simplifier,
            'fpToIEEEBV': self.fptobv_simplifier,
            'fpToFP': self.fptofp_simplifier,
            'StrReverse': self.str_reverse_simplifier,
        }

    def simplify(self, op, args):
        if op not in self._simplifiers:
            return None
        return self._simplifiers[op](*args)

    @staticmethod
    def _deduplicate_filter(args):
        seen = set()
        new_args = []
        for arg in args:
            key = arg.cache_key
            if key in seen:
                continue
            seen.add(key)
            new_args.append(arg)
        return new_args

    #
    # The simplifiers.
    #

    #pylint:disable=inconsistent-return-statements

    @staticmethod
    def if_simplifier(cond, if_true, if_false):
        # NOTE: this is never called; simplifications are implemented inline in the If op. why?
        if cond.is_true():
            return if_true

        if cond.is_false():
            return if_false

    @staticmethod
    def concat_simplifier(*args):

        if len(args) == 1:
            return args[0]

        orig_args = args
        args = list(args)
        #length = sum(arg.length for arg in args)
        simplified = False

        if any(a.symbolic for a in args):
            i = 1
            # here, we concatenate any consecutive concrete terms
            while i < len(args):
                previous = args[i-1]
                current = args[i]

                if not (previous.symbolic or current.symbolic) and backends.concrete.handles(previous) and backends.concrete.handles(current):
                    concatted = ast.all_operations.Concat(previous, current)
                    # If the concrete arguments to concat have non-relocatable annotations attached,
                    # we may not be able to simplify the concrete concat. This check makes sure we don't
                    # create a nested concat in that case.
                    #
                    # This is necessary to ensure that simplified is set correctly. If we don't check this here,
                    # later the concat-of-concat will eliminate the newly introduced concat again, meaning we end
                    # up with the same args that we started with. But because we only eliminate the concat later,
                    # the simplified variable would still be set to True after this loop which is wrong.
                    if concatted.op != "Concat":
                        args[i-1:i+1] = (concatted,)
                        continue

                i += 1

            if len(args) < len(orig_args):
                simplified = True

        # here, we flatten any concats among the arguments and remove zero-length arguments
        i = 0
        while i < len(args):
            current = args[i]
            if current.length == 0:
                args.pop(i)
                simplified = True
            elif current.op == 'Concat':
                simplified = True
                args[i:i+1] = current.args
                i += len(current.args)
            else:
                i += 1

        # here, we consolidate any consecutive concats on extracts from the same variable
        i = 0
        prev_var = None
        prev_left = None
        prev_right = None
        while i < len(args):
            if args[i].op != 'Extract':
                prev_var = None
                prev_left = None
                prev_right = None
                i += 1
            elif prev_var is args[i].args[2] and prev_right == args[i].args[0] + 1:
                prev_right = args[i].args[1]
                args[i-1:i+1] = [ ast.all_operations.Extract(prev_left, prev_right, prev_var) ]
                simplified = True
            else:
                prev_left = args[i].args[0]
                prev_right = args[i].args[1]
                prev_var = args[i].args[2]
                i += 1

        # if any(a.op == 'Reverse' for a in args):
        #	  simplified = True
        #	  args = [a.reversed for a in args]

        if simplified:
            return ast.all_operations.Concat(*args)

        return

    @staticmethod
    def rshift_simplifier(val, shift):
        if (shift == 0).is_true():
            return val
        if val.op == "Concat" and (val.args[0] == 0).is_true() and (shift > val.size() - val.args[0].size()).is_true():
            return ast.all_operations.BVV(0, val.size())
        if val.op == "ZeroExt" and (shift > val.size() - val.args[0]).is_true():
            return ast.all_operations.BVV(0, val.size())

    @staticmethod
    def lshr_simplifier(val, shift):
        if (shift == 0).is_true():
            return val
        if val.op == "Concat" and (val.args[0] == 0).is_true() and (shift > val.size() - val.args[0].size()).is_true():
            return ast.all_operations.BVV(0, val.size())
        if val.op == "ZeroExt" and (shift > val.size() - val.args[0]).is_true():
            return ast.all_operations.BVV(0, val.size())

    @staticmethod
    def lshift_simplifier(val, shift):
        if (shift == 0).is_true():
            return val
        if val.op == '__lshift__':
            real_val, inner_shift = val.args
            return real_val << (inner_shift + shift)

    @staticmethod
    def eq_simplifier(a, b):
        if a is b:
            return ast.true

        if isinstance(a, ast.Bool) and b is ast.true:
            return a
        if isinstance(b, ast.Bool) and a is ast.true:
            return b
        if isinstance(a, ast.Bool) and b is ast.false:
            return ast.all_operations.Not(a)
        if isinstance(b, ast.Bool) and a is ast.false:
            return ast.all_operations.Not(b)

        if a.op == 'Reverse' and b.op == 'Reverse':
            return a.args[0] == b.args[0]

        # simple canonicalization
        if a.op == 'BVV' and b.op != 'BVV':
            return b == a

        # TODO: all these ==/!= might really slow things down...
        if a.op == 'If':
            if a.args[1] is b and ast.all_operations.is_true(a.args[2] != b):
                # (If(c, x, y) == x, x != y) -> c
                return a.args[0]
            elif a.args[2] is b and ast.all_operations.is_true(a.args[1] != b):
                # (If(c, x, y) == y, x != y) -> !c
                return ast.all_operations.Not(a.args[0])
            # elif a._claripy.is_true(a.args[1] == b) and a._claripy.is_true(a.args[2] == b):
            #	  return a._claripy.true
            # elif a._claripy.is_true(a.args[1] != b) and a._claripy.is_true(a.args[2] != b):
            #	  return a._claripy.false

        if b.op == 'If':
            if b.args[1] is a and ast.all_operations.is_true(b.args[2] != b):
                # (x == If(c, x, y)) -> c
                return b.args[0]
            elif b.args[2] is a and ast.all_operations.is_true(b.args[1] != a):
                # (y == If(c, x, y)) -> !c
                return ast.all_operations.Not(b.args[0])
            # elif b._claripy.is_true(b.args[1] == a) and b._claripy.is_true(b.args[2] == a):
            #	  return b._claripy.true
            # elif b._claripy.is_true(b.args[1] != a) and b._claripy.is_true(b.args[2] != a):
            #	  return b._claripy.false

        # Masking and comparing against a constant
        simp = SimplificationManager.and_mask_comparing_against_constant_simplifier(operator.__eq__, a, b)
        if simp is not None:
            return simp

        # ZeroExt/Concat and Extract and comparing against a constant
        simp = SimplificationManager.zeroext_extract_comparing_against_constant_simplifier(operator.__eq__, a, b)
        if simp is not None:
            return simp

        # ZeroExt/Concat and comparing against a constant
        simp = SimplificationManager.zeroext_comparing_against_simplifier(operator.__eq__, a, b)
        if simp is not None:
            return simp

        if (a.op in SIMPLE_OPS or b.op in SIMPLE_OPS) and a.length > 1 and a.length == b.length:
            for i in range(a.length):
                a_bit = a[i:i]
                if a_bit.symbolic:
                    break

                b_bit = b[i:i]
                if b_bit.symbolic:
                    break

                if ast.all_operations.is_false(a_bit == b_bit):
                    return ast.all_operations.false

    @staticmethod
    def ne_simplifier(a, b):
        if a is b:
            return ast.false

        if a.op == 'Reverse' and b.op == 'Reverse':
            return a.args[0] != b.args[0]

        if a.op == 'If':
            if a.args[2] is b and ast.all_operations.is_true(a.args[1] != b):
                # (If(c, x, y) == x, x != y) -> c
                return a.args[0]
            elif a.args[1] is b and ast.all_operations.is_true(a.args[2] != b):
                # (If(c, x, y) == y, x != y) -> !c
                return ast.all_operations.Not(a.args[0])
            # elif a._claripy.is_true(a.args[1] == b) and a._claripy.is_true(a.args[2] == b):
            #	  return a._claripy.false
            # elif a._claripy.is_true(a.args[1] != b) and a._claripy.is_true(a.args[2] != b):
            #	  return a._claripy.true

        if b.op == 'If':
            if b.args[2] is a and ast.all_operations.is_true(b.args[1] != a):
                # (x == If(c, x, y)) -> c
                return b.args[0]
            elif b.args[1] is a and ast.all_operations.is_true(b.args[2] != a):
                # (y == If(c, x, y)) -> !c
                return ast.all_operations.Not(b.args[0])
            # elif b._claripy.is_true(b.args[1] != a) and b._claripy.is_true(b.args[2] != a):
            #	  return b._claripy.true
            # elif b._claripy.is_true(b.args[1] == a) and b._claripy.is_true(b.args[2] == a):
            #	  return b._claripy.false

        # Masking and comparing against a constant
        simp = SimplificationManager.and_mask_comparing_against_constant_simplifier(operator.__ne__, a, b)
        if simp is not None:
            return simp

        # ZeroExt/Concat and Extract and comparing against a constant
        simp = SimplificationManager.zeroext_extract_comparing_against_constant_simplifier(operator.__ne__, a, b)
        if simp is not None:
            return simp

        # ZeroExt/Concat and comparing against a constant
        simp = SimplificationManager.zeroext_comparing_against_simplifier(operator.__ne__, a, b)
        if simp is not None:
            return simp

        if (a.op == SIMPLE_OPS or b.op in SIMPLE_OPS) and a.length > 1 and a.length == b.length:
            for i in range(a.length):
                a_bit = a[i:i]
                if a_bit.symbolic:
                    break

                b_bit = b[i:i]
                if b_bit.symbolic:
                    break

                if ast.all_operations.is_true(a_bit != b_bit):
                    return ast.all_operations.true

    @staticmethod
    def bv_reverse_simplifier(body):
        if body.op == 'Reverse':
            # Reverse(Reverse(x)) ==> x
            return body.args[0]

        if body.length == 8:
            # Reverse(byte) ==> byte
            return body

        if body.op == 'Concat':
            if all(a.op == 'Extract' for a in body.args):
                first_ast = body.args[0].args[2]
                for i,a in enumerate(body.args):
                    if not (first_ast is a.args[2]
                            and a.args[0] == ((i + 1) * 8 - 1)
                            and a.args[1] == i * 8):
                        break
                else:
                    upper_bound = body.args[-1].args[0]
                    if first_ast.length == upper_bound + 1:
                        return first_ast
                    else:
                        return first_ast[upper_bound:0]
            if all(a.length == 8 for a in body.args):
                return body.make_like(body.op, body.args[::-1], simplify=True)

        if body.op == 'Concat':
            if all(a.op == 'Reverse' for a in body.args):
                if all(a.length % 8 == 0 for a in body.args):
                    return body.make_like(body.op, [a.args[0] for a in reversed(body.args)], simplify=True)

        if body.op == 'Extract' and body.args[2].op == 'Reverse':
            # Reverse(Extract(hi, lo, Reverse(x))) ==> Extract(bits-lo-1, bits-hi-1, x)
            # Holds only when (hi+1) and lo are multiples of 8 (or, multiples of bits_per_byte if we really want to
            # suppport cLEMENCy)
            hi, lo = body.args[0:2]
            if (hi + 1) % 8 == 0 and lo % 8 == 0:
                x = body.args[2].args[0]
                new_hi = x.size() - lo - 1
                new_lo = x.size() - hi - 1
                return body.make_like(body.op, (new_hi, new_lo, x), simplify=True)

    @staticmethod
    def boolean_and_simplifier(*args):
        if len(args) == 1:
            return args[0]

        new_args = [None] * len(args)
        ctr = 0
        for a in args:
            if a.op == 'BoolV':
                if a.is_false():
                    return ast.all_operations.false
            else:
                new_args[ctr] = a
                ctr += 1
        new_args = new_args[:ctr]

        if not new_args:
            return ast.true

        if len(new_args) < len(args):
            return ast.all_operations.And(*new_args)

        flattened = SimplificationManager._flatten_simplifier('And', SimplificationManager._deduplicate_filter, *args)
        fargs = flattened.args if flattened is not None else args

        # Check if we are left with one argument again
        if len(fargs) == 1:
            return fargs[0]

        if any(len(arg.args) != 2 for arg in fargs):
            return flattened

        target_var = None

        # Determine the unknown variable
        if fargs[0].args[0].symbolic:
            if fargs[0].args[0] is fargs[1].args[0]:
                target_var = fargs[0].args[0]
            elif fargs[0].args[0] is fargs[1].args[1]:
                target_var = fargs[0].args[0]
        elif fargs[0].args[1].symbolic:
            if fargs[0].args[1] is fargs[1].args[0]:
                target_var = fargs[0].args[1]
            elif fargs[0].args[1] is fargs[1].args[1]:
                target_var = fargs[0].args[1]

        if target_var is None:
            return flattened

        # we now know that the And is a series of binary conditions over a single variable.
        # we can optimize accordingly.
        # right now it's just check for eq/ne

        eq_list = []
        ne_list = []
        for arg in fargs:
            other = arg.args[1] if arg.args[0] is target_var else arg.args[0] if arg.args[1] is target_var else None
            if other is None:
                return flattened
            if arg.op == '__eq__':
                eq_list.append(other)
            elif arg.op == '__ne__':
                ne_list.append(other)
            else:
                return flattened

        if not eq_list:
            return flattened
        if any(any(ne is eq for eq in eq_list) for ne in ne_list):
            return ast.all_operations.false
        if all(v.op == 'BVV' for v in eq_list) and all(v.op == 'BVV' for v in ne_list):
            mustbe = eq_list[0]
            if any(eq.args[0] != mustbe.args[0] for eq in eq_list):
                return ast.all_operations.false
            return target_var == eq_list[0]
        return flattened

    @staticmethod
    def boolean_or_simplifier(*args):
        if len(args) == 1:
            return args[0]

        new_args = []
        for a in args:
            if a.is_true():
                return ast.all_operations.true
            elif not a.is_false():
                new_args.append(a)

        if not new_args:
            return ast.false
        if len(new_args) < len(args):
            return ast.all_operations.Or(*new_args)

        return SimplificationManager._flatten_simplifier('Or', SimplificationManager._deduplicate_filter, *args)

    @staticmethod
    def _flatten_simplifier(op_name, filter_func, *args, **kwargs):
        # we cannot further flatten if any top-level argument has non-relocatable annotations
        if any(not anno.relocatable for anno in itertools.chain.from_iterable(arg.annotations for arg in args)):
            return None

        new_args = tuple(itertools.chain.from_iterable(
            (a.args if isinstance(a, ast.Base) and a.op == op_name else (a,)) for a in args
        ))

        # try to collapse multiple concrete args
        value_args = []
        other_args = []
        for arg in new_args:
            if getattr(arg, 'op', None) == 'BVV':
                value_args.append(arg)
            else:
                other_args.append(arg)
        if other_args and len(value_args) > 1:
            value_arg = value_args[0].make_like(op_name, tuple(value_args), simplify=False)
            new_args = tuple(other_args) + (value_arg,)

        variables = frozenset(itertools.chain.from_iterable(a.variables for a in args if isinstance(a, ast.Base)))
        if filter_func: new_args = filter_func(new_args)
        if not new_args and 'initial_value' in kwargs:
            return kwargs['initial_value']
        # if a single arg is left, don't create an op for it
        if len(new_args) == 1:
            return new_args[0]
        return next(a for a in args if isinstance(a, ast.Base)).make_like(op_name, new_args,
                                                                          variables=variables,
                                                                          simplify=False)

    @staticmethod
    def bitwise_add_simplifier(*args):
        if len(args) == 2 and args[1].op == 'BVV' and args[0].op == '__sub__' and args[0].args[1].op == 'BVV':
            # flatten add over sub
            # (x - y) + z ==> x - (y - z)
            return args[0].args[0] - (args[0].args[1] - args[1])
        return SimplificationManager._flatten_simplifier('__add__', lambda new_args: tuple(a for a in new_args if a.op != 'BVV' or a.args[0] != 0), *args, initial_value=ast.all_operations.BVV(0, len(args[0])))

    @staticmethod
    def bitwise_mul_simplifier(*args):
        return SimplificationManager._flatten_simplifier('__mul__', None, *args)

    @staticmethod
    def bitwise_sub_simplifier(a, b):
        if b.op == 'BVV':
            # many optimizations if b is concrete - effectively flattening
            if b.args[0] == 0:
                return a
            elif a.op == '__sub__' and a.args[1].op == 'BVV':
                # flatten right-heavy trees
                # (x - y) - z ==> x - (y + z)
                return a.args[0] - (a.args[1] + b)
            elif a.op == '__add__' and a.args[-1].op == 'BVV':
                # flatten sub over add
                # (x + y) - z ==> x + (y - z)
                if len(a.args) == 2:
                    return a.args[0] + (a.args[-1] - b)
                else:
                    return a.swap_args(a.args[:-1] + (a.args[-1] - b,))
        elif a is b or (a == b).is_true():
            return ast.all_operations.BVV(0, a.size())
        return None

    # recognize b-bit z=signedmax(q,r) from this idiom:
    # s=q-r;t=q^r;u=s^q;v=u&t;w=v^s;x=rshift(w,b-1);y=x&t;z=q^y
    # and recognize b-bit z=signedmin(q,r) from this idiom:
    # s=r-q;t=q^r;u=s^r;v=u&t;w=v^s;x=rshift(w,b-1);y=x&t;z=q^y
    @staticmethod
    def bitwise_xor_simplifier_minmax(a,b):
        q,y = a,b
        if y.op != '__and__':
            q,y = b,a
            if y.op != '__and__': return

        if len(y.args) != 2: return
        x,t = y.args
        if t.op != '__xor__':
            t,x = y.args
            if t.op != '__xor__': return

        if x.op != '__rshift__': return
        w,dist = x.args

        bits = a.size()
        if dist is not ast.all_operations.BVV(bits - 1,bits): return
        if w.op != '__xor__': return

        if len(w.args) != 2: return
        v,s = w.args
        if s.op != '__sub__':
            s,v = w.args
            if s.op != '__sub__': return

        if v.op != '__and__': return
        if len(v.args) != 2: return
        u,t2 = v.args
        if t2 is not t:
            t2,u = v.args
            if t2 is not t: return

        if u.op != '__xor__': return

        if len(t.args) != 2: return
        q2,r = t.args
        if q2 is not q:
            r,q2 = t.args
            if q2 is not q: return

        if (u.args[0] is s and u.args[1] is q) or (u.args[0] is q and u.args[1] is s):
            if not (s.args[0] is q and s.args[1] is r): return
            cond = ast.all_operations.SLE(q,r)
            return ast.all_operations.If(cond,r,q)

        if (u.args[0] is s and u.args[1] is r) or (u.args[0] is r and u.args[1] is s):
            if not (s.args[0] is r and s.args[1] is q): return
            cond = ast.all_operations.SLE(q,r)
            return ast.all_operations.If(cond,q,r)

    @staticmethod
    def bitwise_xor_simplifier(a, b, *args):
        if not args:
            if a is ast.all_operations.BVV(0, a.size()):
                return b
            elif b is ast.all_operations.BVV(0, a.size()):
                return a
            elif a is b or (a == b).is_true():
                return ast.all_operations.BVV(0, a.size())

            result = SimplificationManager.bitwise_xor_simplifier_minmax(a,b)
            if result is not None: return result

        def _flattening_filter(args):
            # since a ^ a == 0, we can safely remove those from args
            # this procedure is done carefully in order to keep the ordering of arguments
            ctr = collections.Counter(arg.cache_key for arg in args)
            res = []
            seen = set()
            for arg in args:
                if ctr[arg.cache_key] % 2 == 0:
                    continue
                l1 = len(seen)
                seen.add(arg.cache_key)
                l2 = len(seen)
                if l1 != l2:
                    res.append(arg)
            return tuple(res)

        return SimplificationManager._flatten_simplifier('__xor__', _flattening_filter, a, b, *args, initial_value=ast.all_operations.BVV(0, a.size()))

    @staticmethod
    def bitwise_or_simplifier(a, b, *args):
        if not args:
            if a is ast.all_operations.BVV(0, a.size()):
                return b
            if b is ast.all_operations.BVV(0, a.size()):
                return a
            if a.op == b.op and a.op in {"BVV", "BoolV", "FPV"}:
                if a.args == b.args:
                    if (a == b).is_true():
                        return a
            elif (a == b).is_true():
                return a
            if a is b:
                return a

        return SimplificationManager._flatten_simplifier('__or__', SimplificationManager._deduplicate_filter, a, b, *args)

    @staticmethod
    def bitwise_and_simplifier(a, b, *args):
        if not args:
            # try to perform a rotate-shift-mask simplification
            r = SimplificationManager.rotate_shift_mask_simplifier(a, b)
            if r is not None:
                return r
            # we do not use (a == 2 ** a.size()-1).is_true() to avoid creating redundant ASTs
            if a.op == "BVV" and a.args[0] == 2**a.size()-1:
                return b
            if b.op == "BVV" and b.args[0] == 2**b.size()-1:
                return a
            if a is b:
                return a
            # for concrete values, we delay the AST creation as much as possible
            if a.op == b.op and a.op in {"BVV", "BoolV", "FPV"}:
                if a.args == b.args:
                    if (a == b).is_true():
                        return a
            elif (a == b).is_true():
                return a
            if a.op == "BVV" and a.args[0] == 0:
                return ast.all_operations.BVV(0, a.size())
            if b.op == "BVV" and b.args[0] == 0:
                return ast.all_operations.BVV(0, a.size())
            if a.op == "Concat" and len(a.args) == 2:
                # Concat(a.args[0], a.args[1]) & b  ==>  ZeroExt(size, a.args[1])
                # maybe we can drop the second argument
                if (b == 2 ** (a.size() - a.args[0].size()) - 1).is_true():
                    # yes!
                    return ast.all_operations.ZeroExt(a.args[0].size(), a.args[1])

        return SimplificationManager._flatten_simplifier('__and__', SimplificationManager._deduplicate_filter, a, b, *args)

    @staticmethod
    def boolean_not_simplifier(body):
        if body.op == '__eq__':
            return body.args[0] != body.args[1]
        elif body.op == '__ne__':
            return body.args[0] == body.args[1]

        if body.op == 'Not':
            return body.args[0]

        if body.op == 'SLT':
            return ast.all_operations.SGE(body.args[0], body.args[1])
        elif body.op == 'SLE':
            return ast.all_operations.SGT(body.args[0], body.args[1])
        elif body.op == 'SGT':
            return ast.all_operations.SLE(body.args[0], body.args[1])
        elif body.op == 'SGE':
            return ast.all_operations.SLT(body.args[0], body.args[1])

        if body.op == 'ULT':
            return ast.all_operations.UGE(body.args[0], body.args[1])
        elif body.op == 'ULE':
            return ast.all_operations.UGT(body.args[0], body.args[1])
        elif body.op == 'UGT':
            return ast.all_operations.ULE(body.args[0], body.args[1])
        elif body.op == 'UGE':
            return ast.all_operations.ULT(body.args[0], body.args[1])

        if body.op == '__lt__':
            return ast.all_operations.UGE(body.args[0], body.args[1])
        elif body.op == '__le__':
            return ast.all_operations.UGT(body.args[0], body.args[1])
        elif body.op == '__gt__':
            return ast.all_operations.ULE(body.args[0], body.args[1])
        elif body.op == '__ge__':
            return ast.all_operations.ULT(body.args[0], body.args[1])

    @staticmethod
    def zeroext_simplifier(n, e):
        if n == 0:
            return e

        if e.op == 'ZeroExt':
            # ZeroExt(A, ZeroExt(B, x)) ==> ZeroExt(A + B, x)
            return e.make_like(e.op, (n + e.args[0], e.args[1]), length=n + e.size(), simplify=True)

    @staticmethod
    def signext_simplifier(n, e):
        if n == 0:
            return e

        # TODO: if top bit is 0, do a zero-extend instead

    @staticmethod
    def extract_simplifier(high, low, val):
        # if we're extracting the whole value, return the value
        if high - low + 1 == val.size():
            return val

        if (val.op == 'SignExt' or val.op == 'ZeroExt') and low == 0 and high + 1 == val.args[1].size():
            return val.args[1]

        if val.op == 'ZeroExt':
            extending_bits = val.args[0]
            if extending_bits == 0:
                val = val.args[1]
            else:
                val = ast.all_operations.Concat(ast.all_operations.BVV(0, extending_bits), val.args[1])

        # Reverse(concat(a, b)) -> concat(Reverse(b), Reverse(a))
        # a and b must have lengths that are a multiple of 8
        if val.op == 'Reverse' and val.args[0].op == 'Concat' and all(a.length % 8 == 0 for a in val.args[0].args):
            val = ast.all_operations.Concat(*reversed([a.reversed for a in val.args[0].args]))

        # Reading one byte from a reversed ast can be converted to reading the corresponding byte from the original ast
        # No Reverse is required then
        if val.op == 'Reverse' and high - low + 1 == 8 and low % 8 == 0:
            byte_pos = low // 8
            new_byte_pos = val.length // 8 - byte_pos - 1

            val = val.args[0]
            high = (new_byte_pos + 1) * 8 - 1
            low = new_byte_pos * 8

            return ast.all_operations.Extract(high, low, val)

        if val.op == 'Concat':
            pos = val.length
            high_i, low_i, low_loc = None, None, None
            for i, v in enumerate(val.args):
                if pos - v.length <= high < pos:
                    high_i = i
                if pos - v.length <= low < pos:
                    low_i = i
                    low_loc = low - (pos - v.length)
                pos -= v.length

            used = val.args[high_i:low_i+1]
            if len(used) == 1:
                self = used[0]
            else:
                self = ast.all_operations.Concat(*used)

            new_high = low_loc + high - low
            if new_high == self.length - 1 and low_loc == 0:
                return self
            else:
                if self.op != 'Concat':
                    return self[new_high:low_loc]
                else:
                    # to avoid infinite recursion we only return if something was simplified
                    if len(used) != len(val.args) or new_high != high or low_loc != low:
                        return ast.all_operations.Extract(new_high, low_loc, self)

        if val.op == 'Extract':
            _, inner_low = val.args[:2]
            new_low = inner_low + low
            new_high = new_low + (high - low)
            return (val.args[2])[new_high:new_low]

        if val.op == 'Reverse' and val.args[0].op == 'Concat' and all(a.length % 8 == 0 for a in val.args[0].args):
            val = val.make_like('Concat',
                                tuple(reversed([a.reversed for a in val.args[0].args])),
                                simplify=True,
            )[high:low]
            if not val.symbolic:
                return val

        # if all else fails, convert Extract(Reverse(...)) to Reverse(Extract(...))
        # if val.op == 'Reverse' and (high + 1) % 8 == 0 and low % 8 == 0:
        #	  print("saw reverse, converting")
        #	  inner_length = val.args[0].length
        #	  try:
        #		  return val.args[0][(inner_length - 1 - low):(inner_length - 1 - low - (high - low))].reversed
        #	  except ClaripyOperationError:
        #		  __import__('ipdb').set_trace()

        if val.op in extract_distributable:
            all_args = tuple(a[high:low] for a in val.args)
            return reduce(getattr(operator, val.op), all_args)

    # oh gods
    @staticmethod
    def fptobv_simplifier(the_fp):
        if the_fp.op == 'fpToFP' and len(the_fp.args) == 2:
            return the_fp.args[0]

    @staticmethod
    def fptofp_simplifier(*args):
        if len(args) == 2 and args[0].op == 'fpToIEEEBV':
            to_bv, sort = args
            if sort == fp.FSORT_FLOAT and to_bv.length == 32:
                return to_bv.args[0]
            elif sort == fp.FSORT_DOUBLE and to_bv.length == 64:
                return to_bv.args[0]

    @staticmethod
    def rotate_shift_mask_simplifier(a, b):
        """
        Handles the following case:
            ((A << a) | (A >> (_N - a))) & mask, where
                A being a BVS,
                a being a integer that is less than _N,
                _N is either 32 or 64, and
                mask can be evaluated to 0xffffffff (64-bit) or 0xffff (32-bit) after reversing the rotate-shift
                operation.

        It will be simplified to:
            (A & (mask >>> a)) <<< a
        """

        # is the second argument a BVV?
        if b.op != 'BVV':
            return None

        # is it a rotate-shift?
        if a.op != '__or__' or len(a.args) != 2:
            return None
        a_0, a_1 = a.args

        if a_0.op != '__lshift__':
            return None
        if a_1.op != 'LShR':
            return None
        a_00, a_01 = a_0.args
        a_10, a_11 = a_1.args
        if not a_00 is a_10:
            return None
        if a_01.op != 'BVV' or a_11.op != 'BVV':
            return None
        lshift_ = a_01.args[0]
        rshift_ = a_11.args[0]
        bitwidth = lshift_ + rshift_
        if bitwidth not in (32, 64):
            return None

        # is the second argument a mask?
        # Note: the following check can be further loosen if we want to support more masks.
        if bitwidth == 32:
            m = ((b.args[0] << rshift_) & 0xffffffff) | (b.args[0] >> lshift_)
            if m != 0xffff:
                return None
        else: # bitwidth == 64
            m = ((b.args[0] << rshift_) & 0xffffffffffffffff) | (b.args[0] >> lshift_)
            if m != 0xffffffff:
                return None

        # Show our power!
        masked_a = (a_00 & m)
        expr = (masked_a << lshift_) | (masked_a >> rshift_)
        return expr

    @staticmethod
    def str_reverse_simplifier(arg):
        return arg

    @staticmethod
    def and_mask_comparing_against_constant_simplifier(op, a, b):
        """
        This simplifier handles the following case:

            A & mask == b, and
            A & mask != b

        If the high bits of A are 0, `& mask` can be eliminated.
        """
        if op in (operator.__eq__, operator.__ne__) and a.op == '__and__' and len(a.args) == 2 and b.op == "BVV":
            a_arg0, a_arg1 = a.args
            if a_arg1.op == "BVV":
                # it's a constant mask
                # check if the higher bits are 0
                v = a_arg1.args[0]
                zero_bits = a_arg1.args[1]
                mask_allones = True
                while v != 0:
                    mask_allones = mask_allones and ((v & 1) == 0)
                    v = v >> 1
                    zero_bits -= 1
                if zero_bits > 0:
                    # the high `zero_bits` bits are zero
                    # check the constant
                    b_higher = b[b.size() - 1 : b.size() - zero_bits]
                    b_higher_bits_are_0: Optional[bool] = None
                    if (b_higher == 0).is_true():
                        b_higher_bits_are_0 = True
                    elif (b_higher == 0).is_false():
                        b_higher_bits_are_0 = False

                    if b_higher_bits_are_0 is True:
                        # extra check: can we get rid of the mask
                        if zero_bits == b.size():
                            # the mask is 0, which means the left-hand side becomes 0 after masking
                            return op(ast.all_operations.BVV(0, b.size()), b)

                        b_highbit_idx = b.size() - 1 - zero_bits
                        if b.size() % 8 == 0:
                            # originally, b was 8-bit aligned. Can we keep the size of the new expression 8-byte aligned?
                            if (b_highbit_idx + 1) % 8 != 0:
                                b_highbit_idx += 8 - (b_highbit_idx + 1) % 8
                        b_lower = b[b_highbit_idx : 0]
                        if mask_allones:
                            # yes!
                            return op(a_arg0[b_highbit_idx : 0], b_lower)
                        else:
                            # nope
                            if b_highbit_idx == b.size() - 1:
                                # no change is happening
                                return None
                            return op(a_arg0[b_highbit_idx : 0] & a_arg1.args[0], b_lower)
                    elif b_higher_bits_are_0 is False:
                        return ast.all_operations.false if op is operator.__eq__ else ast.all_operations.true

        return None

    @staticmethod
    def zeroext_extract_comparing_against_constant_simplifier(op, a, b):
        """
        This simplifier handles the following cases:

            Extract(hi, 0, Concat(0, A)) op b, and
            Extract(hi, 0, ZeroExt(n, A)) op b

        Extract can be eliminated if the high bits of Concat(0, A) or ZeroExt(n, A) are all zeros.
        """

        # TODO: This method is only used when op is __eq__ or __ne__. We need more experiment to see if it is necessary
        # TODO: to apply this optimization to other comparison operators.

        if a.op != "Extract" or a.args[1] != 0:
            return None

        highbits_are_zeros = False

        a_hi, _, a_arg = a.args
        a_inner_expr = None
        if a_arg.op == "Concat" and len(a_arg.args) == 2:
            a_concat_expr0 = a_arg.args[0]
            a_inner_expr = a_arg.args[1]
            if a_concat_expr0.op == "BVV" and a_concat_expr0.args[0] == 0:
                # equivalent to ZeroExt
                if a_concat_expr0.size() >= a_arg.size() - a_hi:
                    # high bits are all zeros
                    highbits_are_zeros = True

        elif a_arg.op == "ZeroExt":
            a_zeroext_bits = a_arg.args[0]
            a_inner_expr = a_arg.args[1]
            if a_zeroext_bits >= a_arg.size() - a_hi:
                # high bits are all zeros
                highbits_are_zeros = True

        if highbits_are_zeros:
            to_extend = b.size() - a_inner_expr.size()
            if to_extend == 0:
                return op(a_inner_expr, b)
            else:
                return op(ast.all_operations.ZeroExt(to_extend, a_inner_expr), b)

        return None

    @staticmethod
    def zeroext_comparing_against_simplifier(op, a, b):
        """
        This simplifier handles the following cases:

            ZeroExt(n, A) == b, and
            ZeroExt(n, A) != b

        If the high bits of b are all zeros (in case of __eq__) or have at least one ones (in case of __ne__), ZeroExt
        can be eliminated.
        """
        if op in {operator.__eq__, operator.__ne__} and b.op == "BVV":
            # check if the high bits of b are zeros
            if a.op == "ZeroExt":
                a_zeroext_bits = a.args[0]
            elif a.op == "Concat" and len(a.args) == 2 and (a.args[0] == 0).is_true():
                a_zeroext_bits = a.args[0].size()
            else:
                return None

            b_highbits = b[b.size() - 1 : b.size() - a_zeroext_bits]
            if (b_highbits == 0).is_true():
                # we can get rid of ZeroExt
                return op(a.args[1], b[b.size() - a_zeroext_bits - 1 : 0])
            elif (b_highbits == 0).is_false():
                # unsat
                return ast.all_operations.false if op is operator.__eq__ else ast.all_operations.true

        return None


SIMPLE_OPS = ('Concat', 'SignExt', 'ZeroExt')

extract_distributable = {
    '__and__', '__rand__',
    '__or__', '__ror__',
    '__xor__', '__rxor__',
}

from .backend_manager import backends
from . import ast
from . import fp


# the actual instance
simpleton = SimplificationManager()



class SimplifySkipperMixin:
    def __init__(self, *args, **kwargs):
        super(SimplifySkipperMixin, self).__init__(*args, **kwargs)
        self._simplified = True

    def _blank_copy(self, c):
        super(SimplifySkipperMixin, self)._blank_copy(c)
        c._simplified = True

    def _copy(self, c):
        super(SimplifySkipperMixin, self)._copy(c)
        c._simplified = self._simplified

    def __getstate__(self):
        return self._simplified, super().__getstate__()

    def __setstate__(self, s):
        self._simplified, base_state = s
        super().__setstate__(base_state)

    #
    # Simplification skipping
    #

    def add(self, *args, **kwargs):
        added = super(SimplifySkipperMixin, self).add(*args, **kwargs)
        if len(added) > 0:
            self._simplified = False
        return added

    def simplify(self, *args, **kwargs):
        if self._simplified:
            return self.constraints
        else:
            self._simplified = True
            return super(SimplifySkipperMixin, self).simplify(*args, **kwargs)



import os
import pickle
import threading

num_children = 0

import logging
l = logging.getLogger("claripy.backends.backend_z3_parallel")

from .backend_z3 import BackendZ3

class BackendZ3Parallel(BackendZ3):
    def __init__(self):
        BackendZ3.__init__(self)
        self._child = False
        self._lock = threading.RLock()
        self._cache_objects = False

    def _background(self, f_name, *args, **kwargs):
        global num_children

        f = getattr(BackendZ3, f_name)
        if self._child:
            return f(self, *args, **kwargs)

        p_r, p_w = os.pipe()
        p = os.fork()

        if p == 0:
            self._child = True
            self._lock = threading.RLock()
            try:
                r = f(self, *args, **kwargs)
            except UnsatError as e:
                r = e

            # print("WRITING (%d)" % os.getpid())
            pickled = pickle.dumps(r, -1)
            written = 0
            while written < len(pickled):
                written += os.write(p_w, pickled[written:])
            os.close(p_w)
            os.close(p_r)
            # print("WROTE (%d)" % os.getpid())

            os.kill(os.getpid(), 9)
            #os.abort()
            #sys.exit(1)
        else:
            os.close(p_w)

            num_children += 1
            l.debug("in _background with function %s and child %d (among %d)", f, p, num_children)

            # print("READING (from %d)" % p)
            try:
                strs = [ os.read(p_r, 1024*1024) ]
                while strs[-1] != "": strs.append(os.read(p_r, 1024*1024))
            except EOFError:
                raise ClaripyError("read error while receiving data from child")
            os.close(p_r)
            # print("READ (from %d)" % p)

            # thread.start_new_thread(os.wait, ())
            r = pickle.loads("".join(strs))

            os.waitpid(p, 0)
            num_children -= 1
            l.debug("... child %d is done. %d left", p, num_children)

            if isinstance(r, Exception): raise r
            else: return r

    def _synchronize(self, f, *args, **kwargs):
        if self._child:
            return getattr(BackendZ3, f)(self, *args, **kwargs)

        self._lock.acquire()
        #while not self._lock.acquire(blocking=False): print("ACQUIRING...",__import__('time').sleep(1))
        try:
            return getattr(BackendZ3, f)(self, *args, **kwargs)
        finally:
            self._lock.release()

    @staticmethod
    def _translate(*args):
        for a in args:
            if hasattr(a, 'translate'):
                a.translate()

    # backgrounded
    def _results(self, *args, **kwargs):
        return self._background('_results', *args, **kwargs)
    def _eval(self, *args, **kwargs):
        return self._background('_eval', *args, **kwargs)
    def _batch_eval(self, *args, **kwargs):
        return self._background('_batch_eval', *args, **kwargs)
    def _min(self, *args, **kwargs):
        return self._background('_min', *args, **kwargs)
    def _max(self, *args, **kwargs):
        return self._background('_max', *args, **kwargs)

    def _convert(self, *args, **kwargs):
        return self._synchronize('_convert', *args, **kwargs)
    def abstract(self, *args, **kwargs):
        return self._synchronize('abstract', *args, **kwargs)
    def solver(self, *args, **kwargs):
        return self._synchronize('solver', *args, **kwargs)
    def _add(self, *args, **kwargs):
        return self._synchronize('_add', *args, **kwargs)
    def _check(self, *args, **kwargs):
        return self._synchronize('_check', *args, **kwargs)
    def _simplify(self, *args, **kwargs):
        return self._synchronize('_simplify', *args, **kwargs)
    def call(self, *args, **kwargs):
        return self._synchronize('call', *args, **kwargs)
    def resolve(self, *args, **kwargs):
        return self._synchronize('resolve', *args, **kwargs)
    def simplify(self, *args, **kwargs):
        return self._synchronize('simplify', *args, **kwargs)

from ..errors import ClaripyError, UnsatError
